﻿#include "pch.h"
#include "MetroRenderer.h"
#include <DirectXMath.h>
#include <string>
#include <fstream>
#include "DDSTextureLoader.h"
using namespace Microsoft::WRL;
using namespace Windows::Foundation;
using namespace Windows::UI::Core;
using namespace DirectX;
using namespace DX;

MetroRenderer::MetroRenderer() :
    mShapesVB(0), 
	mShapesIB(0), 
	mFloorTexSRV(0),
	mStoneTexSRV(0),
	mBrickTexSRV(0),
	mLightCount(3),
	mMouseDown(false)
{
	mLastMousePos.X = 0;
	mLastMousePos.Y = 0;

	mCam.SetPosition(0.0f, 4.0f, -15.0f);

	XMMATRIX I = XMMatrixIdentity();
	XMStoreFloat4x4(&mGridWorld, I);

	XMMATRIX boxScale = XMMatrixScaling(3.0f, 1.0f, 3.0f);
	XMMATRIX boxOffset = XMMatrixTranslation(0.0f, 0.5f, 0.0f);
	XMStoreFloat4x4(&mBoxWorld, XMMatrixMultiply(boxScale, boxOffset));

	for(int i = 0; i < 5; ++i)
	{
		XMStoreFloat4x4(&mCylWorld[i*2+0], XMMatrixTranslation(-5.0f, 1.5f, -10.0f + i*5.0f));
		XMStoreFloat4x4(&mCylWorld[i*2+1], XMMatrixTranslation(+5.0f, 1.5f, -10.0f + i*5.0f));

		XMStoreFloat4x4(&mSphereWorld[i*2+0], XMMatrixTranslation(-5.0f, 3.5f, -10.0f + i*5.0f));
		XMStoreFloat4x4(&mSphereWorld[i*2+1], XMMatrixTranslation(+5.0f, 3.5f, -10.0f + i*5.0f));
	}

	mDirLights[0].Ambient  = XMFLOAT4(0.2f, 0.2f, 0.2f, 1.0f);
	mDirLights[0].Diffuse  = XMFLOAT4(0.5f, 0.5f, 0.5f, 1.0f);
	mDirLights[0].Specular = XMFLOAT4(0.5f, 0.5f, 0.5f, 1.0f);
	mDirLights[0].Direction = XMFLOAT3(0.57735f, -0.57735f, 0.57735f);

	mDirLights[1].Ambient  = XMFLOAT4(0.0f, 0.0f, 0.0f, 1.0f);
	mDirLights[1].Diffuse  = XMFLOAT4(0.20f, 0.20f, 0.20f, 1.0f);
	mDirLights[1].Specular = XMFLOAT4(0.25f, 0.25f, 0.25f, 1.0f);
	mDirLights[1].Direction = XMFLOAT3(-0.57735f, -0.57735f, 0.57735f);

	mDirLights[2].Ambient  = XMFLOAT4(0.0f, 0.0f, 0.0f, 1.0f);
	mDirLights[2].Diffuse  = XMFLOAT4(0.2f, 0.2f, 0.2f, 1.0f);
	mDirLights[2].Specular = XMFLOAT4(0.0f, 0.0f, 0.0f, 1.0f);
	mDirLights[2].Direction = XMFLOAT3(0.0f, -0.707f, -0.707f);

	mGridMat.Ambient  = XMFLOAT4(0.8f, 0.8f, 0.8f, 1.0f);
	mGridMat.Diffuse  = XMFLOAT4(0.8f, 0.8f, 0.8f, 1.0f);
	mGridMat.Specular = XMFLOAT4(0.8f, 0.8f, 0.8f, 16.0f);

	mCylinderMat.Ambient  = XMFLOAT4(1.0f, 1.0f, 1.0f, 1.0f);
	mCylinderMat.Diffuse  = XMFLOAT4(1.0f, 1.0f, 1.0f, 1.0f);
	mCylinderMat.Specular = XMFLOAT4(0.8f, 0.8f, 0.8f, 16.0f);

	mSphereMat.Ambient  = XMFLOAT4(0.6f, 0.8f, 0.9f, 1.0f);
	mSphereMat.Diffuse  = XMFLOAT4(0.6f, 0.8f, 0.9f, 1.0f);
	mSphereMat.Specular = XMFLOAT4(0.9f, 0.9f, 0.9f, 16.0f);

	mBoxMat.Ambient  = XMFLOAT4(1.0f, 1.0f, 1.0f, 1.0f);
	mBoxMat.Diffuse  = XMFLOAT4(1.0f, 1.0f, 1.0f, 1.0f);
	mBoxMat.Specular = XMFLOAT4(0.8f, 0.8f, 0.8f, 16.0f);
}

void MetroRenderer::CreateDeviceResources()
{
    Direct3DBase::CreateDeviceResources();

	m_window->PointerPressed += ref new TypedEventHandler<CoreWindow^, PointerEventArgs^>(this, &MetroRenderer::OnPointerPressed);
	m_window->PointerMoved   += ref new TypedEventHandler<CoreWindow^, PointerEventArgs^>(this, &MetroRenderer::OnPointerMoved);
	m_window->PointerReleased += ref new TypedEventHandler<CoreWindow^, PointerEventArgs^>(this, &MetroRenderer::OnPointerReleased);
	
	auto loadVSTask = ReadDataAsync("BasicBaseVS.cso");
	auto loadPSTask = ReadDataAsync("Basic3LightFogPS.cso");

	auto createVSTask = loadVSTask.then([this](Platform::Array<byte>^ fileData)
	{
		ThrowIfFailed(m_d3dDevice->CreateVertexShader(fileData->Data, fileData->Length,
				nullptr, &mBasicVS));
		
		const D3D11_INPUT_ELEMENT_DESC vertexDesc[] = 
		{
			{"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0},
			{"NORMAL",   0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0},
			{"TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D11_INPUT_PER_VERTEX_DATA, 0}
		};

		ThrowIfFailed(m_d3dDevice->CreateInputLayout(vertexDesc, ARRAYSIZE(vertexDesc), 
			fileData->Data, fileData->Length, &mInputLayout));
	});

	auto createPSTask = loadPSTask.then([this](Platform::Array<byte>^ fileData) 
	{
		ThrowIfFailed(m_d3dDevice->CreatePixelShader(fileData->Data,
			fileData->Length, nullptr, &mBasicLight3FogPS));
	});

	auto createCubeTask = (createPSTask && createVSTask).then([this] () 
	{
		mPerFrameCB.Initialize(m_d3dDevice.Get());
		mPerObjectCB.Initialize(m_d3dDevice.Get());

		ID3D11Resource* tex = 0;

		ThrowIfFailed(CreateDDSTextureFromFile(m_d3dDevice.Get(), 
			L"Media/Textures/floor.dds", &tex, &mFloorTexSRV));

		// SRV keeps reference.
		ReleaseCOM(tex);

		ThrowIfFailed(CreateDDSTextureFromFile(m_d3dDevice.Get(), 
			L"Media/Textures/stone.dds", &tex, &mStoneTexSRV));

		// SRV keeps reference.
		ReleaseCOM(tex);

		ThrowIfFailed(CreateDDSTextureFromFile(m_d3dDevice.Get(), 
			L"Media/Textures/bricks.dds", &tex, &mBrickTexSRV));
		
		// SRV keeps reference.
		ReleaseCOM(tex);

		D3D11_SAMPLER_DESC samDesc;
		samDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		samDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
		samDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
		samDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
		samDesc.MipLODBias = 0;
		samDesc.MaxAnisotropy = 1;
		samDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		samDesc.BorderColor[ 0 ] = 1.0f;
		samDesc.BorderColor[ 1 ] = 1.0f;
		samDesc.BorderColor[ 2 ] = 1.0f;
		samDesc.BorderColor[ 3 ] = 1.0f;
		samDesc.MinLOD = -3.402823466e+38F; // -FLT_MAX
		samDesc.MaxLOD = 3.402823466e+38F; // FLT_MAX

		ThrowIfFailed(m_d3dDevice->CreateSamplerState(&samDesc, &mSamLinear));

		BuildShapeGeometryBuffers();
		m_loadingComplete = true;

		m_window->PointerCursor = ref new CoreCursor(CoreCursorType::Arrow, 0);
	});
}

void MetroRenderer::CreateWindowSizeDependentResources()
{
    Direct3DBase::CreateWindowSizeDependentResources();

	float aspectRatio = m_windowBounds.Width / m_windowBounds.Height;

	mCam.SetLens(0.25f*MathHelper::Pi, aspectRatio, 1.0f, 1000.0f);
}

void MetroRenderer::Update(float timeTotal, float timeDelta)
{
 
}

void MetroRenderer::Render()
{
    m_d3dContext->OMSetRenderTargets(1, m_renderTargetView.GetAddressOf(), m_depthStencilView.Get());

    const float midnightBlue[] = { 0.098f, 0.098f, 0.439f, 1.000f };
    m_d3dContext->ClearRenderTargetView(m_renderTargetView.Get(), midnightBlue);
    m_d3dContext->ClearDepthStencilView(m_depthStencilView.Get(), D3D11_CLEAR_DEPTH, 1.0f, 0);

    // only draw the cube once it's loaded (this is async)
    if (!m_loadingComplete)
        return;

	//
	// Update the per-frame constant buffer.
	//
	mCam.UpdateViewMatrix();

	XMMATRIX view = mCam.View();
	XMMATRIX proj = mCam.Proj();
	XMMATRIX viewProj = mCam.ViewProj();
	
	XMStoreFloat4x4(&mPerFrameCB.Data.View, XMMatrixTranspose(view));
	XMStoreFloat4x4(&mPerFrameCB.Data.InvView, XMMatrixTranspose(XMMatrixInverse(&XMMatrixDeterminant(view), view)));
    XMStoreFloat4x4(&mPerFrameCB.Data.Proj, XMMatrixTranspose(proj));
	XMStoreFloat4x4(&mPerFrameCB.Data.InvProj, XMMatrixTranspose(XMMatrixInverse(&XMMatrixDeterminant(proj), proj)));
	XMStoreFloat4x4(&mPerFrameCB.Data.ViewProj, XMMatrixTranspose(viewProj));

	mPerFrameCB.Data.DirLights[0] = mDirLights[0];
	mPerFrameCB.Data.DirLights[1] = mDirLights[1];
	mPerFrameCB.Data.DirLights[2] = mDirLights[2];
	mPerFrameCB.Data.EyePosW = mCam.GetPosition();

	mPerFrameCB.Data.gFogStart = 10.0f;
	mPerFrameCB.Data.gFogRange = 60.0f;
	mPerFrameCB.Data.gFogColor = XMFLOAT4(0.65f, 0.65f, 0.65f, 1.0f);

	mPerFrameCB.ApplyChanges(m_d3dContext.Get());

	//
	// Setup IA stage.
	//
	UINT stride = sizeof(VertexBasic32);
    UINT offset = 0;

	m_d3dContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
	m_d3dContext->IASetInputLayout(mInputLayout.Get());
	m_d3dContext->IASetVertexBuffers(0, 1, mShapesVB.GetAddressOf(), &stride, &offset);
	m_d3dContext->IASetIndexBuffer(mShapesIB.Get(), DXGI_FORMAT_R32_UINT, 0);

	//
	// Bind constant buffers.
	//
	ID3D11Buffer* cbuffers[2] = {mPerFrameCB.Buffer(), mPerObjectCB.Buffer()};
	m_d3dContext->VSSetConstantBuffers(0, 2, cbuffers);
	m_d3dContext->PSSetConstantBuffers(0, 2, cbuffers);
	
	//
	// Set sampler.
	//
	m_d3dContext->PSSetSamplers(0, 1, mSamLinear.GetAddressOf());

	//
	// Bind vertex and pixel shaders.
	//
	m_d3dContext->VSSetShader(mBasicVS.Get(), 0, 0);
	m_d3dContext->PSSetShader(mBasicLight3FogPS.Get(), 0, 0);

	//
	// Draw the grid.
	//
	XMMATRIX world = XMLoadFloat4x4(&mGridWorld);
	XMMATRIX worldInvTranspose = MathHelper::InverseTranspose(world);
	XMMATRIX texTransform = XMMatrixScaling(6.0f, 8.0f, 1.0f);

	XMStoreFloat4x4(&mPerObjectCB.Data.World, XMMatrixTranspose(world));
	XMStoreFloat4x4(&mPerObjectCB.Data.WorldInvTranspose, XMMatrixTranspose(worldInvTranspose));
	XMStoreFloat4x4(&mPerObjectCB.Data.TexTransform, XMMatrixTranspose(texTransform));
	mPerObjectCB.Data.Mat = mGridMat;
	mPerObjectCB.ApplyChanges(m_d3dContext.Get());

	m_d3dContext->PSSetShaderResources(0, 1, mFloorTexSRV.GetAddressOf());

	m_d3dContext->DrawIndexed(mGridIndexCount, mGridIndexOffset, mGridVertexOffset);

	//
	// Draw the box.
	//
	world = XMLoadFloat4x4(&mBoxWorld);
	worldInvTranspose = MathHelper::InverseTranspose(world);
	texTransform = XMMatrixIdentity();

	XMStoreFloat4x4(&mPerObjectCB.Data.World, XMMatrixTranspose(world));
	XMStoreFloat4x4(&mPerObjectCB.Data.WorldInvTranspose, XMMatrixTranspose(worldInvTranspose));
	XMStoreFloat4x4(&mPerObjectCB.Data.TexTransform, XMMatrixTranspose(texTransform));
	mPerObjectCB.Data.Mat = mBoxMat;
	mPerObjectCB.ApplyChanges(m_d3dContext.Get());

	m_d3dContext->PSSetShaderResources(0, 1, mStoneTexSRV.GetAddressOf());

	m_d3dContext->DrawIndexed(mBoxIndexCount, mBoxIndexOffset, mBoxVertexOffset);

	//
	// Draw the cylinders.
	//
	for(int i = 0; i < 10; ++i)
	{
		world = XMLoadFloat4x4(&mCylWorld[i]);
		worldInvTranspose = MathHelper::InverseTranspose(world);
		texTransform = XMMatrixIdentity();

		XMStoreFloat4x4(&mPerObjectCB.Data.World, XMMatrixTranspose(world));
		XMStoreFloat4x4(&mPerObjectCB.Data.WorldInvTranspose, XMMatrixTranspose(worldInvTranspose));
		XMStoreFloat4x4(&mPerObjectCB.Data.TexTransform, XMMatrixTranspose(texTransform));
		mPerObjectCB.Data.Mat = mCylinderMat;
		mPerObjectCB.ApplyChanges(m_d3dContext.Get());

		m_d3dContext->PSSetShaderResources(0, 1, mBrickTexSRV.GetAddressOf());

		m_d3dContext->DrawIndexed(mCylinderIndexCount, mCylinderIndexOffset, mCylinderVertexOffset);
	}

	//
	// Draw the spheres.
	//
	for(int i = 0; i < 10; ++i)
	{
		world = XMLoadFloat4x4(&mSphereWorld[i]);
		worldInvTranspose = MathHelper::InverseTranspose(world);
		texTransform = XMMatrixIdentity();

		XMStoreFloat4x4(&mPerObjectCB.Data.World, XMMatrixTranspose(world));
		XMStoreFloat4x4(&mPerObjectCB.Data.WorldInvTranspose, XMMatrixTranspose(worldInvTranspose));
		XMStoreFloat4x4(&mPerObjectCB.Data.TexTransform, XMMatrixTranspose(texTransform));
		mPerObjectCB.Data.Mat = mSphereMat;
		mPerObjectCB.ApplyChanges(m_d3dContext.Get());

		m_d3dContext->PSSetShaderResources(0, 1, mStoneTexSRV.GetAddressOf());

		m_d3dContext->DrawIndexed(mSphereIndexCount, mSphereIndexOffset, mSphereVertexOffset);
	}
}

void MetroRenderer::BuildShapeGeometryBuffers()
{
	GeometryGenerator::MeshData box;
	GeometryGenerator::MeshData grid;
	GeometryGenerator::MeshData sphere;
	GeometryGenerator::MeshData cylinder;

	GeometryGenerator geoGen;
	geoGen.CreateBox(1.0f, 1.0f, 1.0f, 3, box);
	geoGen.CreateGrid(20.0f, 30.0f, 60, 40, grid);
	geoGen.CreateSphere(0.5f, 20, 20, sphere);
	geoGen.CreateCylinder(0.5f, 0.3f, 3.0f, 20, 20, cylinder);

	// Cache the vertex offsets to each object in the concatenated vertex buffer.
	mBoxVertexOffset      = 0;
	mGridVertexOffset     = box.Vertices.size();
	mSphereVertexOffset   = mGridVertexOffset + grid.Vertices.size();
	mCylinderVertexOffset = mSphereVertexOffset + sphere.Vertices.size();

	// Cache the index count of each object.
	mBoxIndexCount      = box.Indices.size();
	mGridIndexCount     = grid.Indices.size();
	mSphereIndexCount   = sphere.Indices.size();
	mCylinderIndexCount = cylinder.Indices.size();

	// Cache the starting index for each object in the concatenated index buffer.
	mBoxIndexOffset      = 0;
	mGridIndexOffset     = mBoxIndexCount;
	mSphereIndexOffset   = mGridIndexOffset + mGridIndexCount;
	mCylinderIndexOffset = mSphereIndexOffset + mSphereIndexCount;
	
	UINT totalVertexCount = 
		box.Vertices.size() + 
		grid.Vertices.size() + 
		sphere.Vertices.size() +
		cylinder.Vertices.size();

	UINT totalIndexCount = 
		mBoxIndexCount + 
		mGridIndexCount + 
		mSphereIndexCount +
		mCylinderIndexCount;

	//
	// Extract the vertex elements we are interested in and pack the
	// vertices of all the meshes into one vertex buffer.
	//

	std::vector<VertexBasic32> vertices(totalVertexCount);

	UINT k = 0;
	for(size_t i = 0; i < box.Vertices.size(); ++i, ++k)
	{
		vertices[k].Pos    = box.Vertices[i].Position;
		vertices[k].Normal = box.Vertices[i].Normal;
		vertices[k].Tex    = box.Vertices[i].TexC;
	}

	for(size_t i = 0; i < grid.Vertices.size(); ++i, ++k)
	{
		vertices[k].Pos    = grid.Vertices[i].Position;
		vertices[k].Normal = grid.Vertices[i].Normal;
		vertices[k].Tex    = grid.Vertices[i].TexC;
	}

	for(size_t i = 0; i < sphere.Vertices.size(); ++i, ++k)
	{
		vertices[k].Pos    = sphere.Vertices[i].Position;
		vertices[k].Normal = sphere.Vertices[i].Normal;
		vertices[k].Tex    = sphere.Vertices[i].TexC;
	}

	for(size_t i = 0; i < cylinder.Vertices.size(); ++i, ++k)
	{
		vertices[k].Pos    = cylinder.Vertices[i].Position;
		vertices[k].Normal = cylinder.Vertices[i].Normal;
		vertices[k].Tex    = cylinder.Vertices[i].TexC;
	}

    D3D11_BUFFER_DESC vbd;
    vbd.Usage = D3D11_USAGE_IMMUTABLE;
    vbd.ByteWidth = sizeof(VertexBasic32) * totalVertexCount;
    vbd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    vbd.CPUAccessFlags = 0;
    vbd.MiscFlags = 0;
    D3D11_SUBRESOURCE_DATA vinitData;
    vinitData.pSysMem = &vertices[0];
    ThrowIfFailed(m_d3dDevice->CreateBuffer(&vbd, &vinitData, &mShapesVB));

	//
	// Pack the indices of all the meshes into one index buffer.
	//

	std::vector<UINT> indices;
	indices.insert(indices.end(), box.Indices.begin(), box.Indices.end());
	indices.insert(indices.end(), grid.Indices.begin(), grid.Indices.end());
	indices.insert(indices.end(), sphere.Indices.begin(), sphere.Indices.end());
	indices.insert(indices.end(), cylinder.Indices.begin(), cylinder.Indices.end());

	D3D11_BUFFER_DESC ibd;
    ibd.Usage = D3D11_USAGE_IMMUTABLE;
    ibd.ByteWidth = sizeof(UINT) * totalIndexCount;
    ibd.BindFlags = D3D11_BIND_INDEX_BUFFER;
    ibd.CPUAccessFlags = 0;
    ibd.MiscFlags = 0;
    D3D11_SUBRESOURCE_DATA iinitData;
    iinitData.pSysMem = &indices[0];
    ThrowIfFailed(m_d3dDevice->CreateBuffer(&ibd, &iinitData, &mShapesIB));
}
 
void MetroRenderer::OnPointerPressed(CoreWindow^ sender, PointerEventArgs^ args)
{
	mLastMousePos = args->CurrentPoint->Position;

	m_window->SetPointerCapture();

	mMouseDown = true;
}

void MetroRenderer::OnPointerMoved(CoreWindow^ sender, PointerEventArgs^ args)
{
	Point currPoint = args->CurrentPoint->Position;

	if(args->CurrentPoint->Properties->IsLeftButtonPressed)
	{
		// Make each pixel correspond to a quarter of a degree.
		float dx = XMConvertToRadians(0.25f*static_cast<float>(currPoint.X - mLastMousePos.X));
		float dy = XMConvertToRadians(0.25f*static_cast<float>(currPoint.Y - mLastMousePos.Y));

		mCam.Pitch(dy);
		mCam.RotateY(dx);
	}
	else if(args->CurrentPoint->Properties->IsRightButtonPressed)
	{
		float dx = 0.02f*static_cast<float>(currPoint.X - mLastMousePos.X);
		float dy = 0.02f*static_cast<float>(currPoint.Y - mLastMousePos.Y);

		mCam.Walk(dy);
		mCam.Strafe(dx);
	}

	mLastMousePos = currPoint;
}

void MetroRenderer::OnPointerReleased(CoreWindow^ sender, PointerEventArgs^ args)
{
	mMouseDown = false;

	m_window->ReleasePointerCapture();
}

