﻿#pragma once

#include <wrl.h>
#include <memory>

#include "Direct3DBase.h"
#include <DirectXMath.h>
#include "GeometryGenerator.h"
#include "MathHelper.h"
#include "LightHelper.h"
#include "Camera.h"
#include "ConstantBuffer.h"

struct PerFrameCB
{
    DirectX::XMFLOAT4X4 View;
	DirectX::XMFLOAT4X4 InvView;
    DirectX::XMFLOAT4X4 Proj;
	DirectX::XMFLOAT4X4 InvProj;
	DirectX::XMFLOAT4X4 ViewProj;

	DirectionalLight DirLights[3];
	DirectX::XMFLOAT3 EyePosW;
	float pad;
	
	DirectX::XMFLOAT4 gFogColor;
	float  gFogStart;
	float  gFogRange;
};

struct PerObjectCB
{
	DirectX::XMFLOAT4X4 World;
	DirectX::XMFLOAT4X4 WorldInvTranspose;
	DirectX::XMFLOAT4X4 TexTransform;
	Material Mat;
};

struct VertexBasic32
{
	DirectX::XMFLOAT3 Pos;
	DirectX::XMFLOAT3 Normal;
	DirectX::XMFLOAT2 Tex;
};

ref class MetroRenderer sealed : public Direct3DBase
{
public:
	MetroRenderer();
	virtual void CreateDeviceResources() override;
	virtual void CreateWindowSizeDependentResources() override;
	virtual void Render() override;
	void Update(float timeTotal, float timeDelta);

private:
	void BuildShapeGeometryBuffers();

	// Pointer event handlers.  We should design app for touch first instead of traditional
	// mouse/keyboard so that it works on tablets.
	void OnPointerPressed(Windows::UI::Core::CoreWindow^ sender, Windows::UI::Core::PointerEventArgs^ args);
	void OnPointerMoved(Windows::UI::Core::CoreWindow^ sender, Windows::UI::Core::PointerEventArgs^ args);
	void OnPointerReleased(Windows::UI::Core::CoreWindow^ sender, Windows::UI::Core::PointerEventArgs^ args);

private:
	bool m_loadingComplete;

	Microsoft::WRL::ComPtr<ID3D11Buffer> mShapesVB;
	Microsoft::WRL::ComPtr<ID3D11Buffer> mShapesIB;

	Microsoft::WRL::ComPtr<ID3D11InputLayout> mInputLayout;

	Microsoft::WRL::ComPtr<ID3D11SamplerState> mSamLinear;

	Microsoft::WRL::ComPtr<ID3D11VertexShader> mBasicVS;
	Microsoft::WRL::ComPtr<ID3D11PixelShader> mBasicLight3FogPS;

	ConstantBuffer<PerFrameCB> mPerFrameCB;
	ConstantBuffer<PerObjectCB> mPerObjectCB;

	Microsoft::WRL::ComPtr<ID3D11ShaderResourceView> mFloorTexSRV;
	Microsoft::WRL::ComPtr<ID3D11ShaderResourceView> mStoneTexSRV;
	Microsoft::WRL::ComPtr<ID3D11ShaderResourceView> mBrickTexSRV;
 
	DirectionalLight mDirLights[3];
	Material mGridMat;
	Material mBoxMat;
	Material mCylinderMat;
	Material mSphereMat;
 
	// Define transformations from local spaces to world space.
	DirectX::XMFLOAT4X4 mSphereWorld[10];
	DirectX::XMFLOAT4X4 mCylWorld[10];
	DirectX::XMFLOAT4X4 mBoxWorld;
	DirectX::XMFLOAT4X4 mGridWorld;

	int mBoxVertexOffset;
	int mGridVertexOffset;
	int mSphereVertexOffset;
	int mCylinderVertexOffset;

	UINT mBoxIndexOffset;
	UINT mGridIndexOffset;
	UINT mSphereIndexOffset;
	UINT mCylinderIndexOffset;

	UINT mBoxIndexCount;
	UINT mGridIndexCount;
	UINT mSphereIndexCount;
	UINT mCylinderIndexCount;

	UINT mLightCount;

	Camera mCam;

	Windows::Foundation::Point mLastMousePos;
	bool mMouseDown;
};
